@@ -1,12 +1,14 @@
 """SDK for building singer-compliant taps."""
 
 from singer_sdk.streams.core import Stream
-from singer_sdk.streams.rest import RESTStream
 from singer_sdk.streams.graphql import GraphQLStream
-
+from singer_sdk.streams.rest import RESTStream
+from singer_sdk.streams.sql import SQLConnector, SQLStream
 
 __all__ = [
     "Stream",
-    "RESTStream",
     "GraphQLStream",
+    "RESTStream",
+    "SQLStream",
+    "SQLConnector",
 ]
