@@ -5,23 +5,27 @@ import copy
 import json
 import sys
 import time
+from io import FileIO
+from pathlib import Path, PurePath
+from typing import IO, Callable, Counter, Dict, List, Optional, Tuple, Type, Union
 
 import click
+from joblib import Parallel, delayed, parallel_backend
 
-from joblib import Parallel, parallel_backend, delayed
-from typing import Any, Dict, Iterable, Optional, Type, List
-
+from singer_sdk.cli import common_options
 from singer_sdk.exceptions import RecordsWitoutSchemaException
 from singer_sdk.helpers._classproperty import classproperty
 from singer_sdk.helpers._compat import final
+from singer_sdk.helpers.capabilities import CapabilitiesEnum, PluginCapabilities
+from singer_sdk.io_base import SingerMessageType, SingerReader
+from singer_sdk.mapper import PluginMapper
 from singer_sdk.plugin_base import PluginBase
 from singer_sdk.sinks import Sink
-from singer_sdk.mapper import PluginMapper
 
 _MAX_PARALLELISM = 8
 
 
-class Target(PluginBase, metaclass=abc.ABCMeta):
+class Target(PluginBase, SingerReader, metaclass=abc.ABCMeta):
     """Abstract base class for targets.
 
     The `Target` class manages config information and is responsible for processing the
@@ -39,16 +43,31 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
 
     def __init__(
         self,
-        config: Optional[Dict[str, Any]] = None,
+        config: Optional[Union[dict, PurePath, str, List[Union[PurePath, str]]]] = None,
         parse_env_config: bool = False,
+        validate_config: bool = True,
     ) -> None:
-        """Initialize the target."""
-        super().__init__(config=config, parse_env_config=parse_env_config)
+        """Initialize the target.
+
+        Args:
+            config: Target configuration. Can be a dictionary, a single path to a
+                configuration file, or a list of paths to multiple configuration
+                files.
+            parse_env_config: Whether to look for configuration values in environment
+                variables.
+            validate_config: True to require validation of config settings.
+        """
+        super().__init__(
+            config=config,
+            parse_env_config=parse_env_config,
+            validate_config=validate_config,
+        )
 
         self._latest_state: Dict[str, dict] = {}
         self._drained_state: Dict[str, dict] = {}
         self._sinks_active: Dict[str, Sink] = {}
         self._sinks_to_clear: List[Sink] = []
+        self._max_parallelism: Optional[int] = _MAX_PARALLELISM
 
         # Approximated for max record age enforcement
         self._last_full_drain_at: float = time.time()
@@ -61,15 +80,43 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         )
 
     @classproperty
-    def capabilities(self) -> List[str]:
-        """Return a list of supported capabilities."""
-        return ["target"]
+    def capabilities(self) -> List[CapabilitiesEnum]:
+        """Get target capabilities.
+
+        Returns:
+            A list of capabilities supported by this target.
+        """
+        return [
+            PluginCapabilities.ABOUT,
+            PluginCapabilities.STREAM_MAPS,
+            PluginCapabilities.FLATTENING,
+        ]
 
     @property
     def max_parallelism(self) -> int:
-        """Return max number of sinks that can be drained in parallel."""
+        """Get max parallel sinks.
+
+        The default is 8 if not overridden.
+
+        Returns:
+            Max number of sinks that can be drained in parallel.
+        """
+        if self._max_parallelism is not None:
+            return self._max_parallelism
+
         return _MAX_PARALLELISM
 
+    @max_parallelism.setter
+    def max_parallelism(self, new_value: int) -> None:
+        """Override the default (max) parallelism.
+
+        The default is 8 if not overridden.
+
+        Args:
+            new_value: The new max degree of parallelism for this target.
+        """
+        self._max_parallelism = new_value
+
     def get_sink(
         self,
         stream_name: str,
@@ -88,8 +135,17 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         sink depending on the values within the `record` object. Otherwise, please see
         `default_sink_class` property and/or the `get_sink_class()` method.
 
-        :raises: RecordsWitoutSchemaException if sink does not exist and schema is not
-                 sent.
+        Raises :class:`singer_sdk.exceptions.RecordsWitoutSchemaException` if sink does
+        not exist and schema is not sent.
+
+        Args:
+            stream_name: Name of the stream.
+            record: Record being processed.
+            schema: Stream schema.
+            key_properties: Primary key of the stream.
+
+        Returns:
+            The sink used for this target.
         """
         _ = record  # Custom implementations may use record in sink selection.
         if schema is None:
@@ -114,12 +170,19 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         return existing_sink
 
     def get_sink_class(self, stream_name: str) -> Type[Sink]:
-        """Return a sink for the given stream name.
+        """Get sink for a stream.
 
         Developers can override this method to return a custom Sink type depending
         on the value of `stream_name`. Optional when `default_sink_class` is set.
 
-        :raises: ValueError if no Sink class is defined.
+        Args:
+            stream_name: Name of the stream.
+
+        Raises:
+            ValueError: If no :class:`singer_sdk.sinks.Sink` class is defined.
+
+        Returns:
+            The sink class to be used with the stream.
         """
         if self.default_sink_class:
             return self.default_sink_class
@@ -130,20 +193,18 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         )
 
     def sink_exists(self, stream_name: str) -> bool:
-        """Return True if a sink has been initialized.
+        """Check sink for a stream.
 
         This method is internal to the SDK and should not need to be overridden.
+
+        Args:
+            stream_name: Name of the stream
+
+        Returns:
+            True if a sink has been initialized.
         """
         return stream_name in self._sinks_active
 
-    @final
-    def listen(self) -> None:
-        """Read from STDIN until all messages are processed.
-
-        This method is internal to the SDK and should not need to be overridden.
-        """
-        self._process_lines(sys.stdin)
-
     @final
     def add_sink(
         self, stream_name: str, schema: dict, key_properties: Optional[List[str]] = None
@@ -151,6 +212,14 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         """Create a sink and register it.
 
         This method is internal to the SDK and should not need to be overridden.
+
+        Args:
+            stream_name: Name of the stream.
+            schema: Schema of the stream.
+            key_properties: Primary key of the stream.
+
+        Returns:
+            A new sink for the stream.
         """
         self.logger.info(f"Initializing '{self.name}' target sink...")
         sink_class = self.get_sink_class(stream_name=stream_name)
@@ -163,16 +232,15 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         self._sinks_active[stream_name] = result
         return result
 
-    @staticmethod
-    def _assert_line_requires(line_dict: dict, requires: List[str]) -> None:
-        for required in requires:
-            if required not in line_dict:
-                raise Exception(
-                    f"Line is missing required '{required}' key: {line_dict}"
-                )
+    def _assert_sink_exists(self, stream_name: str) -> None:
+        """Raise a RecordsWitoutSchemaException exception if stream doesn't exist.
 
-    def _assert_sink_exists(self, stream_name) -> None:
-        """Raise a RecordsWitoutSchemaException exception if stream doesn't exist."""
+        Args:
+            stream_name: TODO
+
+        Raises:
+            RecordsWitoutSchemaException: If sink does not exist and schema is not sent.
+        """
         if not self.sink_exists(stream_name):
             raise RecordsWitoutSchemaException(
                 f"A record for stream '{stream_name}' was encountered before a "
@@ -181,51 +249,39 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
 
     # Message handling
 
-    def _process_lines(self, lines: Iterable[str], table_cache=None) -> None:
+    def _process_lines(self, file_input: IO[str]) -> Counter[SingerMessageType]:
+        """Internal method to process jsonl lines from a Singer tap.
+
+        Args:
+            file_input: Readable stream of messages, each on a separate line.
+
+        Returns:
+            A counter object for the processed lines.
+        """
         self.logger.info(f"Target '{self.name}' is listening for input from tap.")
-        line_counter = 0
-        record_counter = 0
-        state_counter = 0
-        for line in lines:
-            line_counter += 1
-            try:
-                line_dict = json.loads(line)
-            except json.decoder.JSONDecodeError:
-                self.logger.error("Unable to parse:\n{}".format(line))
-                raise
+        counter = super()._process_lines(file_input)
 
-            self._assert_line_requires(line_dict, requires=["type"])
+        line_count = sum(counter.values())
 
-            record_type = line_dict["type"]
-            if record_type == "SCHEMA":
-                self._process_schema_message(line_dict)
-                continue
-
-            if record_type == "RECORD":
-                self._process_record_message(line_dict)
-                record_counter += 1
-                continue
-
-            if record_type == "ACTIVATE_VERSION":
-                self._process_activate_version_message(line_dict)
-                continue
-
-            if record_type == "STATE":
-                self._process_state_message(line_dict)
-                state_counter += 1
-                continue
-
-            raise Exception(f"Unknown message type '{record_type}' in message.")
-
-        self.drain_all()
         self.logger.info(
-            f"Target '{self.name}' completed after {line_counter} lines of input "
-            f"({record_counter} records, {state_counter} state messages)."
+            f"Target '{self.name}' completed reading {line_count} lines of input "
+            f"({counter[SingerMessageType.RECORD]} records, "
+            f"{counter[SingerMessageType.STATE]} state messages)."
         )
 
+        return counter
+
+    def _process_endofpipe(self) -> None:
+        """Called after all input lines have been read."""
+        self.drain_all()
+
     def _process_record_message(self, message_dict: dict) -> None:
-        """Process a RECORD message."""
-        self._assert_line_requires(message_dict, requires=["stream", "record"])
+        """Process a RECORD message.
+
+        Args:
+            message_dict: TODO
+        """
+        self._assert_line_requires(message_dict, requires={"stream", "record"})
 
         stream_name = message_dict["stream"]
         for stream_map in self.mapper.stream_maps[stream_name]:
@@ -259,18 +315,46 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
                 self.drain_one(sink)
 
     def _process_schema_message(self, message_dict: dict) -> None:
-        """Process a SCHEMA messages."""
-        self._assert_line_requires(message_dict, requires=["stream", "schema"])
+        """Process a SCHEMA messages.
+
+        Args:
+            message_dict: The newly received schema message.
+        """
+        self._assert_line_requires(message_dict, requires={"stream", "schema"})
 
         stream_name = message_dict["stream"]
         schema = message_dict["schema"]
         key_properties = message_dict.get("key_properties", None)
+        do_registration = False
         if stream_name not in self.mapper.stream_maps:
-            self.mapper.register_raw_stream_schema(
-                stream_name,
-                schema,
-                key_properties,
+            do_registration = True
+        elif self.mapper.stream_maps[stream_name][0].raw_schema != schema:
+            self.logger.info(
+                f"Schema has changed for stream '{stream_name}'. "
+                "Mapping definitions will be reset."
             )
+            do_registration = True
+        elif (
+            self.mapper.stream_maps[stream_name][0].raw_key_properties != key_properties
+        ):
+            self.logger.info(
+                f"Key properties have changed for stream '{stream_name}'. "
+                "Mapping definitions will be reset."
+            )
+            do_registration = True
+
+        if not do_registration:
+            self.logger.debug(
+                f"No changes detected in SCHEMA message for stream '{stream_name}'. "
+                "Ignoring."
+            )
+            return
+
+        self.mapper.register_raw_stream_schema(
+            stream_name,
+            schema,
+            key_properties,
+        )
         for stream_map in self.mapper.stream_maps[stream_name]:
             # new_schema = helpers._float_to_decimal(new_schema)
             _ = self.get_sink(
@@ -290,8 +374,11 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         """Process a state message. drain sinks if needed.
 
         If state is unchanged, no actions will be taken.
+
+        Args:
+            message_dict: TODO
         """
-        self._assert_line_requires(message_dict, requires=["value"])
+        self._assert_line_requires(message_dict, requires={"value"})
         state = message_dict["value"]
         if self._latest_state == state:
             return
@@ -304,11 +391,14 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
             self.drain_all()
 
     def _process_activate_version_message(self, message_dict: dict) -> None:
-        """Handle the optional ACTIVATE_VERSION message extension."""
-        self.logger.warning(
-            "ACTIVATE_VERSION message received but not supported. Ignoring."
-            "For more information: https://gitlab.com/meltano/sdk/-/issues/18"
-        )
+        """Handle the optional ACTIVATE_VERSION message extension.
+
+        Args:
+            message_dict: TODO
+        """
+        stream_name = message_dict["stream"]
+        sink = self.get_sink(stream_name)
+        sink.activate_version(message_dict["version"])
 
     # Sink drain methods
 
@@ -330,6 +420,9 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         """Drain a specific sink.
 
         This method is internal to the SDK and should not need to be overridden.
+
+        Args:
+            sink: Sink to be drained.
         """
         if sink.current_size == 0:
             return
@@ -339,46 +432,113 @@ class Target(PluginBase, metaclass=abc.ABCMeta):
         sink.mark_drained()
 
     def _drain_all(self, sink_list: List[Sink], parallelism: int) -> None:
-        def _drain_sink(sink: Sink):
+        if parallelism == 1:
+            for sink in sink_list:
+                self.drain_one(sink)
+            return
+
+        def _drain_sink(sink: Sink) -> None:
             self.drain_one(sink)
 
         with parallel_backend("threading", n_jobs=parallelism):
             Parallel()(delayed(_drain_sink)(sink=sink) for sink in sink_list)
 
-    def _write_state_message(self, state: dict):
-        """Emit the stream's latest state."""
+    def _write_state_message(self, state: dict) -> None:
+        """Emit the stream's latest state.
+
+        Args:
+            state: TODO
+        """
         state_json = json.dumps(state)
         self.logger.info(f"Emitting completed target state {state_json}")
-        sys.stdout.write("{}\n".format(state_json))
+        sys.stdout.write(f"{state_json}\n")
         sys.stdout.flush()
 
     # CLI handler
 
     @classproperty
-    def cli(cls):
-        """Execute standard CLI handler for taps."""
+    def cli(cls) -> Callable:
+        """Execute standard CLI handler for taps.
 
-        @click.option("--version", is_flag=True)
-        @click.option("--about", is_flag=True)
-        @click.option("--format")
-        @click.option("--config")
-        @click.command()
+        Returns:
+            A callable CLI object.
+        """
+
+        @common_options.PLUGIN_VERSION
+        @common_options.PLUGIN_ABOUT
+        @common_options.PLUGIN_ABOUT_FORMAT
+        @common_options.PLUGIN_CONFIG
+        @common_options.PLUGIN_FILE_INPUT
+        @click.command(
+            help="Execute the Singer target.",
+            context_settings={"help_option_names": ["--help"]},
+        )
         def cli(
             version: bool = False,
             about: bool = False,
-            config: str = None,
+            config: Tuple[str, ...] = (),
             format: str = None,
-        ):
-            """Handle command line execution."""
+            file_input: FileIO = None,
+        ) -> None:
+            """Handle command line execution.
+
+            Args:
+                version: Display the package version.
+                about: Display package metadata and settings.
+                format: Specify output style for `--about`.
+                config: Configuration file location or 'ENV' to use environment
+                    variables. Accepts multiple inputs as a tuple.
+                file_input: Specify a path to an input file to read messages from.
+                    Defaults to standard in if unspecified.
+
+            Raises:
+                FileNotFoundError: If the config file does not exist.
+            """
             if version:
                 cls.print_version()
                 return
 
-            if about:
-                cls.print_about(format)
-                return
+            if not about:
+                cls.print_version(print_fn=cls.logger.info)
 
-            target = cls(config=config)  # type: ignore  # Ignore 'type not callable'
-            target.listen()
+            validate_config: bool = True
+            if about:
+                validate_config = False
+
+            cls.print_version(print_fn=cls.logger.info)
+
+            parse_env_config = False
+            config_files: List[PurePath] = []
+            for config_path in config:
+                if config_path == "ENV":
+                    # Allow parse from env vars:
+                    parse_env_config = True
+                    continue
+
+                # Validate config file paths before adding to list
+                if not Path(config_path).is_file():
+                    raise FileNotFoundError(
+                        f"Could not locate config file at '{config_path}'."
+                        "Please check that the file exists."
+                    )
+
+                config_files.append(Path(config_path))
+
+            target = cls(  # type: ignore  # Ignore 'type not callable'
+                config=config_files or None,
+                parse_env_config=parse_env_config,
+                validate_config=validate_config,
+            )
+
+            if about:
+                target.print_about(format)
+            else:
+                target.listen(file_input)
 
         return cli
+
+
+class SQLTarget(Target):
+    """Target implementation for SQL destinations."""
+
+    pass
